/* -*- c++ -*- */
/* 
 * Copyright 2015 Ingo D. Rullhusen <d01c@freeshell.de>.
 * with concept from redsea/Oona Räisänen <windyoona@gmail.com>
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include <gnuradio/filter/firdes.h>

#include <volk/volk.h>

#include "clock_impl.h"


namespace gr {
  namespace rds {

    clock::sptr
    clock::make(float samplerate)
    {
      return gnuradio::get_initial_sptr
        (new clock_impl(samplerate));
    }

    /*
     * The private constructor
     */
    static int ios[] = {sizeof(gr_complex), sizeof(float)};
    static std::vector<int> iosig(ios, ios+sizeof(ios)/sizeof(int));
     
    clock_impl::clock_impl(float samplerate)
      : gr::block("clock",
              gr::io_signature::make(1, 1, sizeof(float)),
              gr::io_signature::makev(2, 2, iosig))
    {
       double l, Q, a0, b1, b2;
       std::vector<double> fftaps, fbtaps;

       samp_rate = samplerate;

       d_c_phase = 2.0*M_PI * 57000.0 / samp_rate;


       const double F57 = 57000.0;
       const double B57 =  3500.0;

       fftaps.resize(3), fbtaps.resize(3);

       Q = F57 / B57;

       l = 1.0 / tan( M_PI * F57/samp_rate );
       a0 = ( l * 1.0 / Q ) / ( 1.0 + l/Q + l*l );
       b1 = 2.0*( 1.0 - l*l ) / ( 1.0 + l/Q + l*l );
       b2 = ( 1.0 - l/Q + l*l ) / ( 1.0 + l/Q + l*l );

       fftaps[0] = a0;
       fftaps[1] = 0.0;
       fftaps[2] = -a0;
 
       fbtaps[0] = 1.0;
       fbtaps[1] = b1;
       fbtaps[2] = b2;

       bp_in = new filter::kernel::iir_filter<float,float,double,double>(fftaps, fbtaps, false);


       const double Fg = 2200.0;

       fftaps.resize(2), fbtaps.resize(2);

       const double f1a1 = 1.0;

       l = 1.0 / tan( M_PI * Fg/samp_rate );
       a0 =  1.0 / ( 1.0 + f1a1*l );
       b1 = ( 1.0 - f1a1*l ) / ( 1.0 + f1a1*l );

       fftaps[0] = a0;
       fftaps[1] = a0;
 
       fbtaps[0] = 1.0;
       fbtaps[1] = b1;

       pll = new filter::kernel::iir_filter<float,float,double,double>(fftaps, fbtaps, false);


       const double Flp = 2000.0;

       fftaps.resize(3), fbtaps.resize(3);
       
       const double f2a1 = 1.4142136;
       const double f2b1 = 1.0000000;

       l = 1.0 / tan( M_PI * Flp/samp_rate );
       a0 = 1.0 / ( 1.0 + f2a1*l + f2b1*l*l );
       b1 = 2.0*( 1.0 - f2b1*l*l ) / ( 1.0 + f2a1*l + f2b1*l*l );
       b2 = ( 1.0 - f2a1*l + f2b1*l*l ) / ( 1.0 + f2a1*l + f2b1*l*l );

       fftaps[0] = a0;
       fftaps[1] = 2.0*a0;
       fftaps[2] = a0;
 
       fbtaps[0] = 1.0;
       fbtaps[1] = b1;
       fbtaps[2] = b2;

       lp_out = new filter::kernel::iir_filter<gr_complex,gr_complex,double,gr_complexd>(fftaps, fbtaps, false);

    }

    /*
     * Our virtual destructor.
     */
    clock_impl::~clock_impl()
    {
       delete bp_in;
       delete pll;
       delete lp_out;
    }

    void
    clock_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
        /* <+forecast+> e.g. ninput_items_required[0] = noutput_items */
        ninput_items_required[0] = noutput_items;
    }

    int
    clock_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const float *fmdemod = (const float *) input_items[0];
       gr_complex *out = (gr_complex *) output_items[0];
       float *clk = (float *) output_items[1];

       static double pll_phase = 0;
       static double clock_phase = 0, clock_offset = 0;

       static float prev = 0;

       float f1, f2;
       double c_sin, c_cos;
       gr_complex bb;

       if ( ninput_items[0] < noutput_items ) noutput_items = ninput_items[0];

// Do <+signal processing+>
      for ( int n=0; n<noutput_items; ++n )
      {
         pll_phase += d_c_phase;
         sincos( pll_phase, &c_sin, &c_cos );
 
         f1 = bp_in->filter( fmdemod[n] );
         bb = lp_out->filter( gr_complex( f1 * c_cos, f1 * c_sin ) );
         f2 = pll->filter( bb.real() * bb.imag() );

         pll_phase -= 8.0 * f2;
         d_c_phase -= 4.0 * f2 * 2.0*M_PI/samp_rate;

         out[n] = bb;
 
// 1187.5 Hz clock
         clock_phase = pll_phase / 48.0 + clock_offset;
         clk[n] = (fmod(clock_phase, 2*M_PI) < M_PI ? 1 : -1);

         if ( prev*bb.real() < 0.0 )
         {
            double d_cphi = fmod(clock_phase, M_PI);
            if ( d_cphi >= M_PI_2 ) { d_cphi -= M_PI; }
            clock_offset -= 0.005 * d_cphi;
         }

         prev = bb.real();
      }
      pll_phase = fmod( pll_phase, 2.0*M_PI * 48.0 );

// Tell runtime system how many input items we consumed on
// each input stream.
        consume_each (noutput_items);

// Tell runtime system how many output items we produced.
        return noutput_items;
    }

  } /* namespace rds */
} /* namespace gr */

