/* -*- c++ -*- */
/* 
 * Copyright 2015 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include <gnuradio/prefs.h>
#include "display_impl.h"

#include <QFile>
#include <QTextStream>

#include "crc.h"
#include "rdata.h"

namespace gr
{
   namespace rds
   {

   display::sptr
   display::make( QWidget *parent )
   {
      return gnuradio::get_initial_sptr( new display_impl(parent) );
   }

/*
* The private constructor
*/
   display_impl::display_impl( QWidget *parent )
   : gr::sync_block( "display",
                     gr::io_signature::make(1,1, sizeof(char)),
                     gr::io_signature::make(0, 0, 0)
                   ), d_parent(parent)
   {
// Required now for Qt; argc must be greater than 0 and argv
// must have at least one valid character. Must be valid through
// life of the qApplication:
// http://harmattan-dev.nokia.com/docs/library/html/qt4/qapplication.html
      d_argc = 1;
      d_argv = new char;   
      d_argv[0] = '\0';

       if( qApp != NULL )
      { d_qApplication = qApp; }
      else
      {
#if QT_VERSION >= 0x040500
         std::string style = prefs::singleton()->get_string("qtgui", "style", "raster");
         QApplication::setGraphicsSystem(QString(style.c_str()));
#endif
         d_qApplication = new QApplication(d_argc, &d_argv);
      }

// If a style sheet is set in the prefs file, enable it here.
      std::string qssfile = prefs::singleton()->get_string("qtgui","qss","");
      if( qssfile.size() > 0 )
      {
         QString sstext;
         QFile ss( QString(qssfile.c_str()) );
         if ( ss.open(QIODevice::ReadOnly | QIODevice::Text) )
         {
            QTextStream sstream(&ss);
            while( !sstream.atEnd() )
            { sstext += sstream.readLine(); }
            ss.close();
         }

         d_qApplication->setStyleSheet(sstext);
      }


      d_main_gui = new RDSWidget( d_parent );


   }

/*
* Our virtual destructor.
*/
   display_impl::~display_impl()
   {
      delete d_main_gui;
 
      delete d_argv;
   }


   bool display_impl::check_topology(int ninputs, int noutputs)
   { return( (ninputs == 1) && (noutputs == 0) ); }

   void display_impl::exec_()
   { d_qApplication->exec(); }

   QWidget* display_impl::qwidget()
   { return( d_main_gui ); }

#ifdef ENABLE_PYTHON
   PyObject* display_impl::pyqwidget()
   {
      PyObject *w = PyLong_FromVoidPtr((void*)d_main_gui);
      PyObject *retarg = Py_BuildValue("N", w);
      return retarg;                     
   }
#else
   void* display_impl::pyqwidget()
   { return(NULL); }
#endif

#define rbuffsize 312
#define maxblkcnt 16

   int display_impl::work( int noutput_items,
			   gr_vector_const_void_star &input_items,
			   gr_vector_void_star &output_items )
   {
      const char *in = (const char *) input_items[0];

      static bool sync = false;
      static unsigned char rbuff[ rbuffsize ];
      static int blkerr=0, blkcnt=0;
      static int rb1 = 0, rb2 = 0, grpdetect = -1;
      static U32 blockA, blockB, blockC, blockD;
      static RData D;
      
      int n, bn1, bn2, bn3, bn4, step, err;
      U32 OW;

// Do <+signal processing+>
      for ( int m=0; m<noutput_items; ++m )
      {
         rbuff[rb2] = in[m];
         rb2 = (rb2+1)%rbuffsize;

         n = rb2 - rb1;
         while ( n < 0 ) n += rbuffsize;

         if ( !sync && n > 104 )
         {
            bn1 = blkno( rbuff, rbuffsize, rb1 );
            bn2 = blkno( rbuff, rbuffsize, rb1+26 );
            bn3 = blkno( rbuff, rbuffsize, rb1+52 );
            bn4 = blkno( rbuff, rbuffsize, rb1+78 );

            step = 1;
            if ( bn1==0 && bn2==1 && (bn3==2 || bn3==3) && bn4==4 )
            {
               sync = true; grpdetect=-1;
               step = 0;
            }

            rb1 = (rb1+step)%rbuffsize;
         }

         if ( sync && n > 104 )
         {
            OW = CRC_OW_A;
            err  = decode( rbuff, rbuffsize, rb1,    OW, &blockA );
            OW = CRC_OW_B;
            err += decode( rbuff, rbuffsize, rb1+26, OW, &blockB );
            if ( blockB & 0x00200000 ) { OW = CRC_OW_CS; }
            else                       { OW = CRC_OW_C;  }
            err += decode( rbuff, rbuffsize, rb1+52, OW, &blockC );
            OW = CRC_OW_D;
            err += decode( rbuff, rbuffsize, rb1+78, OW, &blockD );

            blkerr += err;
            blkcnt += 4;

            if ( err < 3 )
            {
               setup_data( &D, blockA, blockB, blockC, blockD );

               if ( blkcnt > maxblkcnt )
               {
                  d_qApplication->postEvent(
                     d_main_gui,
                     new RDSUpdateEvent( D, blkerr/(double)blkcnt )
                                           );
                  blkerr = blkcnt = 0;
               }

               step = 104;
            }
            else
            {
              sync = false;
              step = 1;
           }

            rb1 = (rb1+step)%rbuffsize;
         }

      }

// Tell runtime system how many output items we produced.
      return( noutput_items );
   }

   } /* namespace rds */
} /* namespace gr */

